from uuid import (
    UUID,
)

from minos.common import (
    DatabaseOperation,
)
from minos.saga import (
    SagaExecutionDatabaseOperationFactory,
)

from ...clients import (
    LmdbDatabaseClient,
)
from ...operations import (
    LmdbDatabaseOperation,
    LmdbDatabaseOperationType,
)


class LmdbSagaExecutionDatabaseOperationFactory(SagaExecutionDatabaseOperationFactory):
    """Lmdb Saga Execution Database Operation Factory class."""

    # noinspection PyMethodMayBeStatic
    def build_table_name(self) -> str:
        """Get the table name.

        :return: A ``str`` value.
        """
        return "LocalState"

    def build_store(self, uuid: UUID, **kwargs) -> DatabaseOperation:
        """Build the database operation to store a saga execution.

        :param uuid: The identifier of the saga execution.
        :param kwargs: The attributes of the saga execution.
        :return: A ``DatabaseOperation`` instance.
        """
        key = str(uuid)
        value = kwargs | {"uuid": str(uuid)}

        return LmdbDatabaseOperation(LmdbDatabaseOperationType.CREATE, self.build_table_name(), key, value)

    def build_load(self, uuid: UUID) -> DatabaseOperation:
        """Build the database operation to load a saga execution.

        :param uuid: The identifier of the saga execution.
        :return: A ``DatabaseOperation`` instance.
        """
        key = str(uuid)
        return LmdbDatabaseOperation(LmdbDatabaseOperationType.READ, self.build_table_name(), key)

    def build_delete(self, uuid: UUID) -> DatabaseOperation:
        """Build the database operation to delete a saga execution.

        :param uuid: The identifier of the saga execution.
        :return: A ``DatabaseOperation`` instance.
        """
        key = str(uuid)
        return LmdbDatabaseOperation(LmdbDatabaseOperationType.DELETE, self.build_table_name(), key)


LmdbDatabaseClient.set_factory(SagaExecutionDatabaseOperationFactory, LmdbSagaExecutionDatabaseOperationFactory)



"""This module contains the implementation of the lmdb client."""

from __future__ import (
    annotations,
)

from pathlib import (
    Path,
)
from typing import (
    Any,
    Optional,
    Union,
)

from lmdb import (
    Environment,
)

from minos.common import (
    DatabaseClient,
    DatabaseOperation,
    MinosBinaryProtocol,
    MinosJsonBinaryProtocol,
)

from .operations import (
    LmdbDatabaseOperation,
    LmdbDatabaseOperationType,
)

not_found_sentinel = object()


class LmdbDatabaseClient(DatabaseClient):
    """Lmdb Database Client class."""

    _environment: Optional[Environment]

    def __init__(
        self,
        path: Optional[Union[str, Path]] = None,
        max_tables: int = 100,
        map_size: int = int(1e9),
        protocol: type[MinosBinaryProtocol] = MinosJsonBinaryProtocol,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if path is None:
            path = ".lmdb"

        self._path = path
        self._max_tables = max_tables
        self._map_size = map_size
        self._protocol = protocol
        self._tables = {}

        self._prefetched = not_found_sentinel

        self._environment = None

    async def _setup(self) -> None:
        await super()._setup()
        self._create_environment()

    async def _destroy(self) -> None:
        await super()._destroy()
        self._close_environment()

    def _create_environment(self) -> None:
        self._environment = Environment(str(self._path), max_dbs=self._max_tables, map_size=self._map_size)

    def _close_environment(self) -> None:
        if self._environment is not None:
            self._environment.close()

    async def _reset(self, **kwargs) -> None:
        self._prefetched = None
        self._environment.sync()

    async def _fetch_all(self, *args, **kwargs) -> Any:
        if self._prefetched is not_found_sentinel:
            return
        prefetched = self._prefetched
        self._prefetched = not_found_sentinel
        yield prefetched

    async def _execute(self, operation: DatabaseOperation) -> None:
        if not isinstance(operation, LmdbDatabaseOperation):
            raise ValueError(f"The operation must be a {LmdbDatabaseOperation!r} instance. Obtained: {operation!r}")

        mapper = {
            LmdbDatabaseOperationType.CREATE: self._create,
            LmdbDatabaseOperationType.READ: self._read,
            LmdbDatabaseOperationType.UPDATE: self._update,
            LmdbDatabaseOperationType.DELETE: self._delete,
        }

        fn = mapper[operation.type_]

        fn(table=operation.table, key=operation.key, value=operation.value)

    # noinspection PyUnusedLocal
    def _create(self, table: str, key: str, value: Any, **kwargs) -> None:
        table = self._get_table(table)
        with self._environment.begin(write=True) as transaction:
            encoded = self._protocol.encode(value)
            transaction.put(key.encode(), encoded, db=table)

    # noinspection PyUnusedLocal
    def _read(self, table: str, key: str, **kwargs):
        table = self._get_table(table)
        with self._environment.begin(db=table) as transaction:
            value = transaction.get(key.encode(), default=not_found_sentinel)
            if value is not not_found_sentinel:
                value = self._protocol.decode(value)

        self._prefetched = value

    # noinspection PyUnusedLocal
    def _delete(self, table: str, key: str, **kwargs) -> None:
        table = self._get_table(table)
        with self._environment.begin(write=True, db=table) as transaction:
            transaction.delete(key.encode())

    # noinspection PyUnusedLocal
    def _update(self, table: str, key: str, value: Any, **kwargs) -> None:
        table = self._get_table(table)
        with self._environment.begin(write=True, db=table) as transaction:
            encoded = self._protocol.encode(value)
            transaction.put(key.encode(), encoded, db=table, overwrite=True)

    def _get_table(self, table: str) -> Any:
        if table not in self._tables:
            self._tables[table] = self._environment.open_db(table.encode())
        return self._tables[table]



"""The lmdb plugin of the Minos Framework."""

__author__ = "Minos Framework Devs"
__email__ = "hey@minos.run"
__version__ = "0.8.0.dev1"

from .clients import (
    LmdbDatabaseClient,
)
from .factories import (
    LmdbSagaExecutionDatabaseOperationFactory,
)
from .operations import (
    LmdbDatabaseOperation,
    LmdbDatabaseOperationType,
)



from enum import (
    Enum,
)
from typing import (
    Any,
    Optional,
)

from minos.common import (
    DatabaseOperation,
)


class LmdbDatabaseOperationType(str, Enum):
    """Lmdb Database Operation Type class."""

    CREATE = "create"
    READ = "read"
    UPDATE = "update"
    DELETE = "delete"


class LmdbDatabaseOperation(DatabaseOperation):
    """Lmdb Database Operation class."""

    def __init__(
        self, type_: LmdbDatabaseOperationType, table: str, key: str, value: Optional[Any] = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.type_ = type_
        self.table = table
        self.key = key
        self.value = value
